@@ -176,15 +176,21 @@ class TestBaseAgent(ironic_agent_base.IronicAgentTest):
     @mock.patch(
         'ironic_python_agent.hardware_managers.cna._detect_cna_card',
         mock.Mock())
+    @mock.patch.object(hardware, 'dispatch_to_managers', autospec=True)
     @mock.patch.object(agent.IronicPythonAgent,
                        '_wait_for_interface', autospec=True)
-    @mock.patch.object(hardware, 'dispatch_to_managers', autospec=True)
-    @mock.patch('wsgiref.simple_server.make_server', autospec=True)
-    def test_run(self, mock_make_server, mock_dispatch, mock_wait):
+    @mock.patch('wsgiref.simple_server.WSGIServer', autospec=True)
+    @mock.patch.object(hardware, 'load_managers', autospec=True)
+    def test_run(self, mock_load_managers, mock_wsgi,
+                 mock_wait, mock_dispatch):
         CONF.set_override('inspection_callback_url', '')
-        wsgi_server = mock_make_server.return_value
-        wsgi_server.start.side_effect = KeyboardInterrupt()
 
+        wsgi_server = mock_wsgi.return_value
+
+        def set_serve_api():
+            self.agent.serve_api = False
+
+        wsgi_server.handle_request.side_effect = set_serve_api
         self.agent.heartbeater = mock.Mock()
         self.agent.api_client.lookup_node = mock.Mock()
         self.agent.api_client.lookup_node.return_value = {
@@ -195,21 +201,66 @@ class TestBaseAgent(ironic_agent_base.IronicAgentTest):
                 'heartbeat_timeout': 300
             }
         }
+
         self.agent.run()
 
         listen_addr = agent.Host('192.0.2.1', 9999)
-        mock_make_server.assert_called_once_with(
-            listen_addr.hostname,
-            listen_addr.port,
-            self.agent.api,
-            server_class=simple_server.WSGIServer)
-        wsgi_server.serve_forever.assert_called_once_with()
+        mock_wsgi.assert_called_once_with(
+            (listen_addr.hostname,
+             listen_addr.port),
+            simple_server.WSGIRequestHandler)
+        wsgi_server.set_app.assert_called_once_with(self.agent.api)
+        self.assertTrue(wsgi_server.handle_request.called)
         mock_wait.assert_called_once_with(mock.ANY)
         self.assertEqual([mock.call('list_hardware_info'),
                           mock.call('wait_for_disks')],
                          mock_dispatch.call_args_list)
         self.agent.heartbeater.start.assert_called_once_with()
 
+    @mock.patch.object(hardware, '_check_for_iscsi', mock.Mock())
+    @mock.patch(
+        'ironic_python_agent.hardware_managers.cna._detect_cna_card',
+        mock.Mock())
+    @mock.patch.object(agent.IronicPythonAgent,
+                       '_wait_for_interface', autospec=True)
+    @mock.patch.object(hardware, 'dispatch_to_managers', autospec=True)
+    @mock.patch('wsgiref.simple_server.WSGIServer', autospec=True)
+    @mock.patch.object(hardware, 'load_managers', autospec=True)
+    def test_run_raise_exception(self, mock_load_managers, mock_wsgi,
+                                 mock_dispatch, mock_wait):
+        CONF.set_override('inspection_callback_url', '')
+
+        wsgi_server = mock_wsgi.return_value
+        wsgi_server.handle_request.side_effect = KeyboardInterrupt()
+        self.agent.heartbeater = mock.Mock()
+        self.agent.api_client.lookup_node = mock.Mock()
+        self.agent.api_client.lookup_node.return_value = {
+            'node': {
+                'uuid': 'deadbeef-dabb-ad00-b105-f00d00bab10c'
+            },
+            'config': {
+                'heartbeat_timeout': 300
+            }
+        }
+
+        self.assertRaisesRegex(errors.IronicAPIError,
+                               'Failed due to unknow exception.',
+                               self.agent.run)
+
+        self.assertTrue(mock_wait.called)
+        self.assertEqual([mock.call('list_hardware_info'),
+                          mock.call('wait_for_disks')],
+                         mock_dispatch.call_args_list)
+        listen_addr = agent.Host('192.0.2.1', 9999)
+        mock_wsgi.assert_called_once_with(
+            (listen_addr.hostname,
+             listen_addr.port),
+            simple_server.WSGIRequestHandler)
+        wsgi_server.set_app.assert_called_once_with(self.agent.api)
+        self.assertTrue(wsgi_server.handle_request.called)
+        self.agent.heartbeater.start.assert_called_once_with()
+        self.assertTrue(wsgi_server.handle_request.called)
+
     @mock.patch.object(hardware, '_check_for_iscsi', mock.Mock())
     @mock.patch('ironic_python_agent.hardware_managers.cna._detect_cna_card',
                 mock.Mock())
@@ -217,15 +268,18 @@ class TestBaseAgent(ironic_agent_base.IronicAgentTest):
                        '_wait_for_interface', autospec=True)
     @mock.patch.object(inspector, 'inspect', autospec=True)
     @mock.patch.object(hardware, 'dispatch_to_managers', autospec=True)
-    @mock.patch('wsgiref.simple_server.make_server', autospec=True)
+    @mock.patch('wsgiref.simple_server.WSGIServer', autospec=True)
     @mock.patch.object(hardware.HardwareManager, 'list_hardware_info',
                        autospec=True)
-    def test_run_with_inspection(self, mock_list_hardware, mock_make_server,
+    def test_run_with_inspection(self, mock_list_hardware, mock_wsgi,
                                  mock_dispatch, mock_inspector, mock_wait):
         CONF.set_override('inspection_callback_url', 'http://foo/bar')
 
-        wsgi_server = mock_make_server.return_value
-        wsgi_server.start.side_effect = KeyboardInterrupt()
+        def set_serve_api():
+            self.agent.serve_api = False
+
+        wsgi_server = mock_wsgi.return_value
+        wsgi_server.handle_request.side_effect = set_serve_api
 
         mock_inspector.return_value = 'uuid'
 
@@ -242,12 +296,12 @@ class TestBaseAgent(ironic_agent_base.IronicAgentTest):
         self.agent.run()
 
         listen_addr = agent.Host('192.0.2.1', 9999)
-        mock_make_server.assert_called_once_with(
-            listen_addr.hostname,
-            listen_addr.port,
-            self.agent.api,
-            server_class=simple_server.WSGIServer)
-        wsgi_server.serve_forever.assert_called_once_with()
+        mock_wsgi.assert_called_once_with(
+            (listen_addr.hostname,
+             listen_addr.port),
+            simple_server.WSGIRequestHandler)
+        self.assertTrue(mock_wsgi.called)
+
         mock_inspector.assert_called_once_with()
         self.assertEqual(1, self.agent.api_client.lookup_node.call_count)
         self.assertEqual(
@@ -268,12 +322,12 @@ class TestBaseAgent(ironic_agent_base.IronicAgentTest):
                        '_wait_for_interface', autospec=True)
     @mock.patch.object(inspector, 'inspect', autospec=True)
     @mock.patch.object(hardware, 'dispatch_to_managers', autospec=True)
-    @mock.patch('wsgiref.simple_server.make_server', autospec=True)
+    @mock.patch('wsgiref.simple_server.WSGIServer', autospec=True)
     @mock.patch.object(hardware.HardwareManager, 'list_hardware_info',
                        autospec=True)
     def test_run_with_inspection_without_apiurl(self,
                                                 mock_list_hardware,
-                                                mock_make_server,
+                                                mock_wsgi,
                                                 mock_dispatch,
                                                 mock_inspector,
                                                 mock_wait):
@@ -295,18 +349,21 @@ class TestBaseAgent(ironic_agent_base.IronicAgentTest):
         self.assertFalse(hasattr(self.agent, 'api_client'))
         self.assertFalse(hasattr(self.agent, 'heartbeater'))
 
-        wsgi_server = mock_make_server.return_value
-        wsgi_server.start.side_effect = KeyboardInterrupt()
+        def set_serve_api():
+            self.agent.serve_api = False
+
+        wsgi_server = mock_wsgi.return_value
+        wsgi_server.handle_request.side_effect = set_serve_api
 
         self.agent.run()
 
         listen_addr = agent.Host('192.0.2.1', 9999)
-        mock_make_server.assert_called_once_with(
-            listen_addr.hostname,
-            listen_addr.port,
-            self.agent.api,
-            server_class=simple_server.WSGIServer)
-        wsgi_server.serve_forever.assert_called_once_with()
+        mock_wsgi.assert_called_once_with(
+            (listen_addr.hostname,
+             listen_addr.port),
+            simple_server.WSGIRequestHandler)
+        self.assertTrue(wsgi_server.handle_request.called)
+
         mock_inspector.assert_called_once_with()
 
         self.assertFalse(mock_wait.called)
@@ -320,12 +377,12 @@ class TestBaseAgent(ironic_agent_base.IronicAgentTest):
                        '_wait_for_interface', autospec=True)
     @mock.patch.object(inspector, 'inspect', autospec=True)
     @mock.patch.object(hardware, 'dispatch_to_managers', autospec=True)
-    @mock.patch('wsgiref.simple_server.make_server', autospec=True)
+    @mock.patch('wsgiref.simple_server.WSGIServer', autospec=True)
     @mock.patch.object(hardware.HardwareManager, 'list_hardware_info',
                        autospec=True)
     def test_run_without_inspection_and_apiurl(self,
                                                mock_list_hardware,
-                                               mock_make_server,
+                                               mock_wsgi,
                                                mock_dispatch,
                                                mock_inspector,
                                                mock_wait):
@@ -347,18 +404,20 @@ class TestBaseAgent(ironic_agent_base.IronicAgentTest):
         self.assertFalse(hasattr(self.agent, 'api_client'))
         self.assertFalse(hasattr(self.agent, 'heartbeater'))
 
-        wsgi_server = mock_make_server.return_value
-        wsgi_server.start.side_effect = KeyboardInterrupt()
+        def set_serve_api():
+            self.agent.serve_api = False
+
+        wsgi_server = mock_wsgi.return_value
+        wsgi_server.handle_request.side_effect = set_serve_api
 
         self.agent.run()
 
         listen_addr = agent.Host('192.0.2.1', 9999)
-        mock_make_server.assert_called_once_with(
-            listen_addr.hostname,
-            listen_addr.port,
-            self.agent.api,
-            server_class=simple_server.WSGIServer)
-        wsgi_server.serve_forever.assert_called_once_with()
+        mock_wsgi.assert_called_once_with(
+            (listen_addr.hostname,
+             listen_addr.port),
+            simple_server.WSGIRequestHandler)
+        self.assertTrue(wsgi_server.handle_request.called)
 
         self.assertFalse(mock_inspector.called)
         self.assertFalse(mock_wait.called)
@@ -392,12 +451,16 @@ class TestBaseAgent(ironic_agent_base.IronicAgentTest):
     @mock.patch.object(agent.IronicPythonAgent, '_wait_for_interface',
                        autospec=True)
     @mock.patch.object(hardware, 'dispatch_to_managers', autospec=True)
-    @mock.patch('wsgiref.simple_server.make_server', autospec=True)
+    @mock.patch('wsgiref.simple_server.WSGIServer', autospec=True)
     def test_run_with_sleep(self, mock_make_server, mock_dispatch,
                             mock_wait, mock_sleep, mock_load_managers):
         CONF.set_override('inspection_callback_url', '')
+
+        def set_serve_api():
+            self.agent.serve_api = False
+
         wsgi_server = mock_make_server.return_value
-        wsgi_server.start.side_effect = KeyboardInterrupt()
+        wsgi_server.handle_request.side_effect = set_serve_api
 
         self.agent.hardware_initialization_delay = 10
         self.agent.heartbeater = mock.Mock()
@@ -414,11 +477,9 @@ class TestBaseAgent(ironic_agent_base.IronicAgentTest):
 
         listen_addr = agent.Host('192.0.2.1', 9999)
         mock_make_server.assert_called_once_with(
-            listen_addr.hostname,
-            listen_addr.port,
-            self.agent.api,
-            server_class=simple_server.WSGIServer)
-        wsgi_server.serve_forever.assert_called_once_with()
+            (listen_addr.hostname,
+             listen_addr.port),
+            simple_server.WSGIRequestHandler)
 
         self.agent.heartbeater.start.assert_called_once_with()
         mock_sleep.assert_called_once_with(10)
@@ -504,6 +565,14 @@ class TestBaseAgent(ironic_agent_base.IronicAgentTest):
         source = self.agent._get_route_source('XXX')
         self.assertEqual('1:2::3:4', source)
 
+    @mock.patch.object(utils, 'execute', autospec=True)
+    def test_get_route_source_ipv6_linklocal(self, mock_execute):
+        mock_execute.return_value = (
+            'XXX src fe80::1234:1234:1234:1234 metric XXX\n    cache', None)
+
+        source = self.agent._get_route_source('XXX')
+        self.assertIsNone(source)
+
     @mock.patch.object(agent, 'LOG', autospec=True)
     @mock.patch.object(utils, 'execute', autospec=True)
     def test_get_route_source_indexerror(self, mock_execute, mock_log):
@@ -539,32 +608,35 @@ class TestAgentStandalone(ironic_agent_base.IronicAgentTest):
         'ironic_python_agent.hardware_managers.cna._detect_cna_card',
         mock.Mock())
     @mock.patch('wsgiref.simple_server.make_server', autospec=True)
+    @mock.patch('wsgiref.simple_server.WSGIServer', autospec=True)
     @mock.patch.object(hardware.HardwareManager, 'list_hardware_info',
                        autospec=True)
-    def test_run(self, mock_list_hardware, mock_make_server):
+    @mock.patch.object(hardware, 'load_managers', autospec=True)
+    def test_run(self, mock_load_managers, mock_list_hardware,
+                 mock_wsgi, mock_make_server):
         wsgi_server = mock_make_server.return_value
         wsgi_server.start.side_effect = KeyboardInterrupt()
+        wsgi_server_request = mock_wsgi.return_value
+
+        def set_serve_api():
+            self.agent.serve_api = False
+
+        wsgi_server_request.handle_request.side_effect = set_serve_api
 
         self.agent.heartbeater = mock.Mock()
         self.agent.api_client.lookup_node = mock.Mock()
-        self.agent.api_client.lookup_node.return_value = {
-            'node': {
-                'uuid': 'deadbeef-dabb-ad00-b105-f00d00bab10c'
-            },
-            'config': {
-                'heartbeat_timeout': 300
-            }
-        }
+
         self.agent.run()
 
+        self.assertTrue(mock_load_managers.called)
         listen_addr = agent.Host('192.0.2.1', 9999)
-        mock_make_server.assert_called_once_with(
-            listen_addr.hostname,
-            listen_addr.port,
-            self.agent.api,
-            server_class=simple_server.WSGIServer)
-        wsgi_server.serve_forever.assert_called_once_with()
+        mock_wsgi.assert_called_once_with(
+            (listen_addr.hostname,
+             listen_addr.port),
+            simple_server.WSGIRequestHandler)
+        wsgi_server_request.set_app.assert_called_once_with(self.agent.api)
 
+        self.assertTrue(wsgi_server_request.handle_request.called)
         self.assertFalse(self.agent.heartbeater.called)
         self.assertFalse(self.agent.api_client.lookup_node.called)
 
